-- Pylon: a foundation library; version 0.82(beta); 1998-01-04

-- 
-- Copyright (c) 1997 Franck Arnaud, England, E.U.
-- 
-- Permission is hereby granted, without written agreement and without
-- licence or royalty fees, to use, copy, modify, or distribute this source
-- code or binary programs using it or part of it, subject to the following
-- condition:
-- 
-- The above copyright notice and the following disclaimer must appear in
-- all copies or substantial portions of this source code. (Binary programs
-- derived from the source code can be distributed without restriction or
-- mention of the origin of the source.)
-- 
-- THIS SOFTWARE IS PROVIDED BY FRANCK ARNAUD AND THE AUTHORS ``AS IS'' AND
-- ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS BE LIABLE TO ANY PARTY FOR 
-- ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
-- DAMAGES ARISING IN ANY WAY OUT OF THE USE OF THIS SOURCE CODE.
-- 

indexing
	cluster: Pylon, Container
	description: "Tree"
	interface: client

class P_GENERAL_TREE[G]

inherit
	P_TREE[G]
		end

creation
	make

feature

	make, reset is
		local
			node: P_TREE_NODE[G]
		do
			current_node := Void
			root_node := Void
			count := 0
			make_traversable
			!!node -- SmallEiffel type coverage
		end

feature -- Public attribute(s)
	
	count : INTEGER

feature {P_GENERAL_TREE, P_MUTATOR } -- Attributes
	
	current_node : P_TREE_NODE[G]
	root_node : like current_node

feature {P_MUTATOR }
	
	set_current_node(cn : like current_node) is
		require
			ok: cn /= Void
		do
			current_node := cn
		end

feature {NONE } -- (Internal) Copy

	recursive_subtree_copy(node : P_TREE_NODE[G]) is
		require
			valid: node /= Void
		local
			child : P_TREE_NODE[G]
			first_done : BOOLEAN
			tok : P_ONEWAY_ITERATOR[G]
		do
			
			from
				child := node.first_child
			until
				child = Void
			loop

				if first_done then
					add_right(child.item)
				else
					add_child(child.item)
					first_done := true
				end
				
				tok := preorder_iterator -- save
				tok.synchronize
				tok.lock

				recursive_subtree_copy(child)
				
				tok.unlock
				tok.synchronize_container
				tok.stop -- restore

				check first_ok: first_done end

				child := child.right
			end
		end

feature {NONE } -- (Internal) Count

	reset_count is
		do
			if root_node /= Void then
				count := 1
				recursive_count(root_node)
			else
				count := 0
			end
		end

	recursive_count(node : P_TREE_NODE[G]) is
		require
			valid: node /= Void
		local
			child : P_TREE_NODE[G]
		do
			from
				child := node.first_child
			until
				child = Void
			loop
				count := count + 1
				recursive_count(child)
				
				child := child.right
			end
		end

feature {P_GENERAL_TREE }
	
	copy_from_node(node : like current_node) is
		do
			add_root(node.item)
			recursive_subtree_copy(node)
			check unlocked: not has_locked end
			reset_count
			check unlocked_count: not has_locked end 
		end

feature -- General

	copy(other : like Current) is
		do
			copy_from_node(other.root_node)
		end

	is_equal(other : like Current) : BOOLEAN is
			-- Is this tree equal with other?
			-- Time: O(count)
		local
			it, it_other : P_ONEWAY_ITERATOR[G]
		do
			--? good method?

			-- Tree equal if all items and depths identical during 
			-- a preorder iteration.
			if other.count = count then
				from
					it := preorder_iterator
					it_other := other.preorder_iterator
					Result := true
				invariant
					it.outside = it_other.outside
				variant
					it.forward_variant
				until
					Result = false or it.outside
				loop
					it.synchronize_container
					it_other.synchronize_container
					
					if it.item /= it_other.item or depth /= other.depth then
						Result := false
					end

					-- forth
					it.forth
					it_other.forth
				end
			end
		end

feature -- Container

	item : G is
		do
			Result := current_node.item
		end

	depth : INTEGER is
			-- Number of level from the root. Root depth is 0.
			-- Time: O(depth)
		local
			lnode : P_TREE_NODE[G]
		do
			from
				lnode := current_node
			until
				lnode.parent = Void
			loop
				Result := Result + 1
				lnode := lnode.parent
			end
		end

feature -- Iteration

	preorder_iterator, linear_iterator : P_ONEWAY_ITERATOR[G] is
			-- Preorder iterator on (forest) tree.
		local
			actual : P_TREE_PRE_ITERATOR[G]
		do
			!!actual.make(Current)
			Result := actual
		end

	postorder_iterator : P_ONEWAY_ITERATOR[G] is
			-- Postorder iterator on (forest) tree.
		local
			actual : P_TREE_POST_ITERATOR[G]
		do
			!!actual.make(Current)
			Result := actual
		end

	levelorder_iterator : P_ONEWAY_ITERATOR[G] is
			-- Level order iterator on (forest) tree.
		local
			actual : P_TREE_LEVEL_ITERATOR[G]
		do
			!!actual.make(Current) 
			Result := actual
		end

feature -- Lock status

	is_locked : BOOLEAN is
			-- Is current item locked?
			-- Time: O(1)
		do
			Result := current_node.is_locked
		end

feature -- Cursor
	
	is_left : BOOLEAN is
		do
			Result := (current_node.left = Void)
		end

	is_right : BOOLEAN is
		do
			Result := (current_node.right = Void)
		end

	is_top : BOOLEAN is
		do
			Result := (current_node.parent = Void)
		end

	is_bottom : BOOLEAN is
		do
			Result := (current_node.first_child = Void)
		end

	leftmost is
		do
			from
			until
				current_node.left = Void
			loop
				current_node := current_node.left
			end
		end

	rightmost is
		do
			from
			until
				current_node.right = Void
			loop
				current_node := current_node.right
			end
		end

	right is
			-- Go backwards one slot.
		do
			current_node := current_node.right
		end

	left is
			-- Go forward one item.
		do
			current_node := current_node.left
		end

	up is
			-- Go one position upwards.
		do
			current_node := current_node.parent
		end

	down is
			-- Go to the first item of the next generation.
		do
			current_node := current_node.first_child
		end
	
	root is
			-- Go to root node.
		do
			current_node := root_node
		end

feature -- Modification

	add_left(new : G) is
			-- Add an item to the left of current item.
		local
			new_node : like current_node
		do
			-- make new_node
			!!new_node
			new_node.set_parent(current_node.parent)
			new_node.set_left(current_node.left)
			new_node.set_right(current_node)
			new_node.set_item(new)

			-- insert
			if current_node.left /= Void then
				current_node.left.set_right(new_node)
			else -- current_node.left = Void
				if current_node.parent /= Void then
					current_node.parent.set_first_child(new_node)
				else -- is_top and is_left, move root
					root_node := new_node
				end
			end
			current_node.set_left(new_node)
			
			-- install
			current_node := new_node
			count := count + 1
		end

	add_right(new : G) is
			-- Add an item to the right of current item.
		local
			new_node : like current_node
		do
			-- prepare
			!!new_node
			new_node.set_parent(current_node.parent)
			new_node.set_right(current_node.right)
			new_node.set_left(current_node)
			new_node.set_item(new)

			-- insert
			if current_node.right /= Void then
				current_node.right.set_left(new_node)
			end
			current_node.set_right(new_node)

			-- install
			current_node := new_node
			count := count + 1
		end

	add_root(new_root : G) is
			-- Add the root item.
		do
			check count: count = 0 end
			
			!!root_node
			root_node.set_item(new_root)
			current_node := root_node 
			
			count := 1
		end
	
	add_child(new : G) is
			-- Add a first child to the current item.
			-- Other children may be added with add_right/add_left
		local
			new_node : like current_node
		do
			!!new_node
			new_node.set_item(new)
			new_node.set_parent(current_node)
			
			current_node.set_first_child(new_node)

			-- install
			current_node := new_node
			count := count + 1
		end

	replace(x : G) is
			-- Replace current item.
		do
			current_node.set_item(x)
		end

	remove is
			-- Remove current item and its children (subtree).
			-- The parent becomes current node.
			-- Time: O(count) 
		do
			--? improve recalculation of count
			
			if current_node.first_child /= Void then -- useful?
				current_node.set_first_child(Void)
				-- subtree will be collected.
			end
			if current_node.left /= Void then
				current_node.left.set_right(current_node.right)
			else -- current_node.left = Void
				if current_node.parent /= Void then
					current_node.parent.set_first_child(current_node.right)
				end
			end
			if current_node.right /= Void then
				current_node.right.set_left(current_node.left)
			end
			if current_node.parent /= Void then
				current_node := current_node.parent
			else
				current_node := Void
				root_node := Void
			end

			-- recalculate count
			reset_count
		end

feature -- Extra cursor

	subtree : like Current is
		do
			!!Result.make
			Result.copy_from_node(current_node)
		end

	siblings_list : P_LIST[G] is
			-- List of siblings of current item, including current item.
		local
			node : P_TREE_NODE[G]
			list : P_LINKED_LIST[G]
		do
			!!list.make
			Result := list
			
			-- carriage return
			from
				node := current_node
			until
				node.left = Void
			loop
				node := node.left
			end
		
			-- iterate
			from
			until
				node = Void
			loop
				Result.add(node.item)
				node := node.right
			end
		end

invariant
	valid_current: current_node /= Void implies current_node.is_valid

end
