-- Pylon: a foundation library; version 0.82(beta); 1998-01-04

-- 
-- Copyright (c) 1997 Franck Arnaud, England, E.U.
-- 
-- Permission is hereby granted, without written agreement and without
-- licence or royalty fees, to use, copy, modify, or distribute this source
-- code or binary programs using it or part of it, subject to the following
-- condition:
-- 
-- The above copyright notice and the following disclaimer must appear in
-- all copies or substantial portions of this source code. (Binary programs
-- derived from the source code can be distributed without restriction or
-- mention of the origin of the source.)
-- 
-- THIS SOFTWARE IS PROVIDED BY FRANCK ARNAUD AND THE AUTHORS ``AS IS'' AND
-- ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS BE LIABLE TO ANY PARTY FOR 
-- ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
-- DAMAGES ARISING IN ANY WAY OUT OF THE USE OF THIS SOURCE CODE.
-- 

indexing
	cluster: Pylon, Misc
	description: "Text object support"
	interface: client
	 
class P_TEXT_OBJECT_MANAGER

creation {P_TEXT_OBJECT}
	make_in, make_out

feature {NONE} -- Supplier(s)
	
	fmt : P_FORMAT is
		once --?
			!!Result 
		end

feature {P_TEXT_OBJECT} -- Public interface
	
	make_out is
		do
			!!output_string.make(0)
		end

	make_in(str : STRING) is
		require
			valid: str /= Void
		do
			!!instr.make(0)
			instr.copy(str)
			instr_position := 1
			in_success := true
		end

feature {P_TEXT_OBJECT}
	
	output_string : STRING

feature {NONE}
	
	instr : STRING
	instr_position : INTEGER

feature -- Input

	is_output : BOOLEAN is
			-- Is the manager in output mode?
		do
			Result := output_string /= Void
		end
	
	put_boolean(bool:BOOLEAN) is
			-- Put a boolean value in the output string.
		require
			out: is_output
		do
			if bool = true then
				put_entry(Boolean_id,"true")
			else
				put_entry(Boolean_id,"false")
			end
		end

	put_character(char : CHARACTER) is
			-- Put a character in the output string.
		require
			out: is_output
		local
			str : STRING
			idec : P_INTEGER_DECODER
		do
			!!str.make(0)
			if is_encoded_char(char) then
				!!idec.make(Encoding_base)
				str.append_character(Encoding_character)
				str.append_string(idec.value_to_string(char.code))
			else
				str.append_character(char)
			end
			put_entry(Character_id, str)
		end

	put_integer(int: INTEGER) is
			-- Put an integer in the output string.
		do
			put_entry(Integer_id, fmt.int_to_str(int))
		end

	put_string(outs: STRING) is
			-- Put a data string in the output string.
		require
			out: is_output
			valid_out: outs /= Void
		do
			put_entry(String_id, encode_string(outs))
		end

	put_object(obj : P_TEXT_OBJECT) is
			-- Put an encoded object in the output string.
		require
			out: is_output
			valid_obj: obj /= Void
		do
			put_entry(Textobj_id, obj.to_string)
		end

	put_real(number : REAL) is
			-- Put a real in the output string.
		require
			out: is_output
		do
			put_entry(Real_id, fmt.real_to_str(number))
		end

	put_double(number : DOUBLE) is
			-- Put a double in the output string.
		require
			out: is_output
		do
			put_entry(Double_id, fmt.double_to_str(number))
		end

feature -- Input

	in_success : BOOLEAN
			-- Is the input process without errors?

	is_input : BOOLEAN is
			-- Is the manager in input mode?
		do
			Result := instr /= Void
		end

	get_boolean is
			-- Retrieve boolean value from stream.
		require
			in: is_input
		do
			get_entry(Boolean_id)
			if fmt.case_equal(last_entry,"true") then
				last_boolean := true
			else
				last_boolean := false
			end
		ensure
			has_result: -- last_boolean set
		end

	last_boolean : BOOLEAN

	get_character is
			-- Retrieve character from stream.
		require
			in: is_input
		local
			ch : CHARACTER
			idec : P_INTEGER_DECODER
		do
			get_entry(Character_id)
			if last_entry.count >= 1 then
				ch := last_entry.item(1)
				if ch = Encoding_character and last_entry.count > 1 then
					!!idec.make(Encoding_base)
					idec.from_string(last_entry.substring(2, last_entry.count))
					--? value overflow
					last_character := fmt.code_to_char(idec.value)
				else
					last_character := ch
				end
			end
		ensure
			has_result: -- last_character set.
		end

	last_character : CHARACTER

	get_integer is
			-- Retrieve integer from stream.
		require
			in: is_input
		do
			get_entry(Integer_id)
			last_integer := fmt.str_to_int(last_entry)
		ensure
			has_result: -- last_integer set.
		end

	last_integer : INTEGER

	get_string is
			-- Retrieve string from stream.
		require
			in: is_input
		do
			get_entry(String_id)
			last_string := decode_string(last_entry)
		ensure
			has_result: -- last_string set.
		end

	last_string : STRING

	get_object(obj_to_change : P_TEXT_OBJECT) is
			-- Retrieve encoded object string, and adapt consequently 
			-- object given as parameter.
		require
			in: is_input
		do
			get_entry(Textobj_id)
			if in_success then
				obj_to_change.from_string(last_entry)
			end
		end

	get_real is
			-- Retrieve real number from string.
		require
			in: is_input
		do
			get_entry(Real_id)
			last_real := fmt.str_to_real(last_entry)
		ensure
			has_result: -- last_real set.
		end

	last_real : REAL

	get_double is
			-- Retrieve double precision real number from stream.
		require
			in: is_input
		do
			get_entry(Double_id)
			last_double := fmt.str_to_double(last_entry)
		ensure
			has_result: -- last_double set.
		end

	last_double : DOUBLE

feature {NONE} -- Base items
	
	String_id : STRING is "STR"
	Textobj_id : STRING is "OBJ"
	Integer_id : STRING is "INT"
	Boolean_id : STRING is "BOOL"
	Character_id : STRING is "CHAR"
	Real_id : STRING is "REAL"
	Double_id : STRING is "DBLE"

	is_entry_encoded(str : STRING) : BOOLEAN is
			-- Check that string has no encodable character.
		local
			i, cnt_parenthesis : INTEGER
			ch : CHARACTER
		do
			from
				i := 1
				Result := true
			variant str.count + 1 - i until 
				(not Result) or (i > str.count)
			loop
				ch := str.item(i)
				if ch = Encoding_character then
					-- OK
				elseif ch = '(' then
					cnt_parenthesis := cnt_parenthesis + 1
				elseif ch = ')' then
					cnt_parenthesis := cnt_parenthesis - 1
				else
					Result := not is_encoded_char(ch)
				end
				i := i + 1
			end

			Result := Result and (cnt_parenthesis = 0)
		end

	put_entry(id:STRING; data:STRING) is
			-- Put a low-level entry.
		require
			valid_out: output_string /= Void
			valid_data: data /= Void
			valid_encoding: is_entry_encoded(data)
		do
			output_string.append_string("(")
			output_string.append_string(id)
			output_string.append_string(":")
			output_string.append_string(data)
			output_string.append_string(")")
		end

	get_entry(id:STRING) is
			-- Retrieve a low-level entry.
		local
			i, cnt, inside, start : INTEGER
		do
			-- init
			!!last_entry.make(0)
			cnt := instr.count

			if in_success then
				-- find '('
				if instr_position > cnt or else instr.item(instr_position) /= '(' then
					in_success := false
				else
					-- find end of tag marker (:)
					from
						i := instr_position + 1
					variant cnt + 1 - i until 
						i > cnt or else instr.item(i) = ':' 
					loop
						i := i + 1
					end

					-- check we have found it.
					if i > cnt or else i <= instr_position + 1 then
						in_success := false
					else
						-- check good tag. 
						if not fmt.case_equal(id, instr.substring(instr_position+1, i-1)) then
							in_success := false
						else
								
							-- find entry content field end marker ')'
							-- ignore pairs of parenthesis inside
							from
								start := i + 1
								inside := 0
							until (i > cnt) or else ((inside = 0) and (instr.item(i) = ')')) loop
								if instr.item(i) = '(' then
									inside := inside + 1
								elseif instr.item(i) = ')' then
									inside := inside - 1
								end
								i := i + 1
							end
							
							-- check we found it.
							if i > cnt then
								in_success := false
							else
								-- copy entry.
								if i > start then
									last_entry := instr.substring(start, i-1)
								end
								-- update input position 
								instr_position := i + 1
							end -- )
						end -- tag
					end -- found :
				end -- found (
			end
		end

	last_entry : STRING

feature {NONE} -- Character encoding

	Encoding_character : CHARACTER is '%%'
	Encoding_base : INTEGER is 16

	Min_clear : INTEGER is 32
	Max_clear : INTEGER is 127

	is_encoded_char(ch : CHARACTER) : BOOLEAN is
			-- Should this character be encoded?
		do
			Result := (ch.code < Min_clear) or (ch.code > Max_clear) or (ch = '(') or (ch = ')') or (ch = Encoding_character)
		end

feature {NONE} -- String Encoding.

	encode_string(str : STRING) : STRING is
			-- Encode string, escape <32 and () and Encoding_character
		require
			valid: str /= Void
		local
			i : INTEGER
			ch : CHARACTER
			idec : P_INTEGER_DECODER
		do
			!!idec.make(Encoding_base)
			!!Result.make(0)
			
			from
				i := 1
			until
				i > str.count
			loop
				ch := str.item(i)
				if is_encoded_char(ch) then
					Result.append_character(Encoding_character)
					Result.append_string(idec.value_to_string(ch.code))
					Result.append_character(Encoding_character)
				else
					Result.append_character(ch)
				end
				i := i + 1
			end
		end

	decode_string(str : STRING) : STRING is
			-- Decode %ab12% in string.
		require
			valid: str /= Void
		local
			in_encoded_char : BOOLEAN
			ch : CHARACTER
			i : INTEGER
			idec : P_INTEGER_DECODER
		do
			!!idec.make(Encoding_base)
			!!Result.make(0)

			from
				i := 1
			variant str.count + 1 - i until i > str.count loop
				
				ch := str.item(i)
				
				if in_encoded_char then
					if ch = Encoding_character then
						--? if code too big...
						Result.append_character(fmt.code_to_char(idec.value))
						in_encoded_char := false
					else
						if idec.is_digit(ch) then
							idec.extend(ch)
						end --? else ignore
					end
				else -- not encoded
					if ch = Encoding_character then
						in_encoded_char := true
						idec.reset
					else
						Result.append_character(ch)
					end
				end
				
				i := i + 1
			end
		end

feature {NONE}
	
	remove_string(str : STRING; position : INTEGER; lenght : INTEGER) is
		require
			str /= Void
		local
			i : INTEGER
		do
			from
				i := position
			variant
				lenght + position - i + 1 
			until
				i > position + lenght
			loop
				str.remove(position)
				i := i + 1
			end
		end

end
